# (c) Zygmunt Krynicki 2005, 2006, 2007, 2008
# Licensed under GPL, see COPYING for the whole text

from __future__ import print_function

import gettext
import grp
import json
import os
import os.path
import posix
import sys
import subprocess

from CommandNotFound.db.db import SqliteDatabase

if sys.version >= "3":
    _gettext_method = "gettext"
else:
    _gettext_method = "ugettext"
_ = getattr(gettext.translation("command-not-found", fallback=True), _gettext_method)

    
def similar_words(word):
    """
    return a set with spelling1 distance alternative spellings

    based on http://norvig.com/spell-correct.html
    """
    alphabet = 'abcdefghijklmnopqrstuvwxyz-_0123456789'
    s = [(word[:i], word[i:]) for i in range(len(word) + 1)]
    deletes = [a + b[1:] for a, b in s if b]
    transposes = [a + b[1] + b[0] + b[2:] for a, b in s if len(b) > 1]
    replaces = [a + c + b[1:] for a, b in s for c in alphabet if b]
    inserts = [a + c + b     for a, b in s for c in alphabet]
    return set(deletes + transposes + replaces + inserts)


def user_can_sudo():
    try:
        groups = posix.getgroups()
        return (grp.getgrnam("sudo")[2] in groups or
                grp.getgrnam("admin")[2] in groups)
    except KeyError:
        return False


# the new style DB - if that exists we skip the legacy DB 
dbpath = "/var/lib/command-not-found/commands.db"
# the legacy DB shipped in the command-not-found-data package
legacy_db = "/usr/share/command-not-found/commands.db"
        

class CommandNotFound(object):

    programs_dir = "programs.d"
    max_len = 256
    
    prefixes = (
        "/bin",
        "/usr/bin",
        "/usr/local/bin",
        "/sbin",
        "/usr/sbin",
        "/usr/local/sbin",
        "/usr/games")

    def __init__(self, data_dir="/usr/share/command-not-found"):
        self.sources_list = self._getSourcesList()
        # a new style DB means we can skip loading the old legacy static DB
        if os.path.exists(dbpath):
            self.db = SqliteDatabase(dbpath)
        else:
            self.db = SqliteDatabase(legacy_db)
        self.user_can_sudo = user_can_sudo()

    def spelling_suggestions(self, word, min_len=3):
        """ try to correct the spelling """
        possible_alternatives = []
        if not (min_len <= len(word) <= self.max_len):
            return possible_alternatives
        for w in similar_words(word):
            packages = self.get_packages(w)
            for (package, ver, comp) in packages:
                possible_alternatives.append((w, package, comp, ver))
        return possible_alternatives

    def print_spelling_suggestions(self, word, mispell_packages, mispell_snaps, max_alt=15):
        """ print spelling suggestions for packages and snaps """
        if len(mispell_packages)+len(mispell_snaps) > max_alt:
            print(_("Command '%s' not found, but there are %s similar ones") % (word, len(mispell_packages)), file=sys.stderr)
        elif len(mispell_packages)+len(mispell_snaps) > 0:
            print()
            print(_("Command '%s' not found, did you mean:") % word, file=sys.stderr)
            print()
            for (w, p, ver) in mispell_snaps:
                if ver:
                    ver = " (%s) " % ver
                else:
                    ver = ""
                print(_("  Command '%s' from snap '%s'%s") % (w, p, ver), file=sys.stderr)
            for (w, p, c, ver) in mispell_packages:
                if ver:
                    ver = " (%s, %s)" % (ver, c)
                else:
                    ver = " (%s)" % c
                print(_("  Command '%s' from package '%s'%s") % (w, p, ver), file=sys.stderr)
        print()

    def get_packages(self, command):
        return self.db.lookup(command)

    def get_snaps(self, command):
        exact_result = set()
        mispell_result = set()
        if not os.path.exists("/usr/bin/snap"):
            return set(), set()
        try:
            with open(os.devnull) as devnull:
                output = subprocess.check_output(
                    ["/usr/bin/snap", "advise-snap", "--format=json",
                     "--command", command],
                    stderr=devnull,
                    universal_newlines=True)
        except subprocess.CalledProcessError:
            return set(), set()
        try:
            snaps = json.loads(output)
        except json.JSONDecodeError:
            return set(), set()
        for snap in snaps:
            if snap["Command"] == command:
                exact_result.add((snap["Snap"], snap["Command"], snap.get("Version")))
            else:
                mispell_result.add((snap["Snap"], snap["Command"], snap.get("Version")))
        return exact_result, mispell_result
    
    def getBlacklist(self):
        try:
            with open(os.sep.join((os.getenv("HOME", "/root"), ".command-not-found.blacklist"))) as blacklist:
                return [line.strip() for line in blacklist if line.strip() != ""]
        except IOError:
            return []

    def _getSourcesList(self):
        try:
            import apt_pkg
            from aptsources.sourceslist import SourcesList
            apt_pkg.init()
        except (SystemError, ImportError):
            return []
        sources_list = set([])
        # The matcher parses info files from
        # /usr/share/python-apt/templates/
        # But we don't use the calculated data, skip it
        for source in SourcesList(withMatcher=False):
            if not source.disabled and not source.invalid:
                for component in source.comps:
                    sources_list.add(component)
        return sources_list

    def install_prompt(self, package_name):
        if not "COMMAND_NOT_FOUND_INSTALL_PROMPT" in os.environ:
            return
        if package_name:
            prompt = _("Do you want to install it? (N/y)")
            if sys.version >= '3':
                answer = input(prompt)
                raw_input = lambda x: x  # pyflakes
            else:
                answer = raw_input(prompt)
                if sys.stdin.encoding and isinstance(answer, str):
                    # Decode the answer so that we get an unicode value
                    answer = answer.decode(sys.stdin.encoding)
            if answer.lower() == _("y"):
                if posix.geteuid() == 0:
                    command_prefix = ""
                else:
                    command_prefix = "sudo "
                install_command = "%sapt install %s" % (command_prefix, package_name)
                print("%s" % install_command, file=sys.stdout)
                subprocess.call(install_command.split(), shell=False)

    def advise(self, command, ignore_installed=False):
        " give advice where to find the given command to stderr "
        def _in_prefix(prefix, command):
            " helper that returns if a command is found in the given prefix "
            return (os.path.exists(os.path.join(prefix, command))
                    and not os.path.isdir(os.path.join(prefix, command)))

        if len(command) > self.max_len:
            return False
        
        if command.startswith("/"):
            if os.path.exists(command):
                prefixes = [os.path.dirname(command)]
            else:
                prefixes = []
        else:
            prefixes = [prefix for prefix in self.prefixes if _in_prefix(prefix, command)]

        # check if we have it in a common prefix that may not be in the PATH
        if prefixes and not ignore_installed:
            if len(prefixes) == 1:
                print(_("Command '%(command)s' is available in '%(place)s'") % {"command": command, "place": os.path.join(prefixes[0], command)}, file=sys.stderr)
            else:
                print(_("Command '%(command)s' is available in the following places") % {"command": command}, file=sys.stderr)
                for prefix in prefixes:
                    print(" * %s" % os.path.join(prefix, command), file=sys.stderr)
            missing = list(set(prefixes) - set(os.getenv("PATH", "").split(":")))
            if len(missing) > 0:
                print(_("The command could not be located because '%s' is not included in the PATH environment variable.") % ":".join(missing), file=sys.stderr)
                if "sbin" in ":".join(missing):
                    print(_("This is most likely caused by the lack of administrative privileges associated with your user account."), file=sys.stderr)
            return False

        # do not give advice if we are in a situation where apt
        # or aptitude are not available (LP: #394843)
        if not (os.path.exists("/usr/bin/apt") or
                os.path.exists("/usr/bin/aptitude")):
            return False

        if command in self.getBlacklist():
            return False
        packages = self.get_packages(command)
        snaps, mispell_snaps = self.get_snaps(command)
        if len(packages) == 0 and len(snaps) == 0:
            mispell_packages = self.spelling_suggestions(command)
            self.print_spelling_suggestions(command, mispell_packages, mispell_snaps)
        elif len(snaps) > 0 and len(packages) == 0:
            # use something like jinja (just more lightweight)
            print()
            print(_("Command '%s' not found, but can be installed as:") % command, file=sys.stderr)
            print()
            for snap in snaps:
                ver = ""
                if snap[2]:
                    ver = " (%s) " % snap[2]
                print(" %s%s" % (snap[0], ver), file=sys.stderr)
            print()
            print(_("See 'snap info <snap name>' for additional versions."))
            print()
        elif len(packages) == 1 and len(snaps) == 0:
            print(_("Command '%s' not found, but can be installed as:") % command, file=sys.stderr)
            if posix.geteuid() == 0:
                print("apt install %s" % packages[0][0], file=sys.stderr)
                self.install_prompt(packages[0][0])
            elif self.user_can_sudo:
                print("sudo apt install %s" % packages[0][0], file=sys.stderr)
                self.install_prompt(packages[0][0])
            else:
                print("apt install %s" % packages[0][0], file=sys.stderr)
                print(_("Please ask your administrator."))
                if not packages[0][2] in self.sources_list:
                    print(_("You will have to enable the component called '%s'") % packages[0][2], file=sys.stderr)
        elif len(packages) > 1 and len(snaps) == 0:
            print()
            print(_("Command '%s' not found, but can be installed as:") % command, file=sys.stderr)
            print()
            for package in packages:
                if package[2] in self.sources_list:
                    print(" %s" % package[0], file=sys.stderr)
                else:
                    print(" %s" % package[0] + " (" + _("You will have to enable component called '%s'") % package[2] + ")", file=sys.stderr)
            print()
            if posix.geteuid() == 0:
                print(_("Try: %s <deb name>") % "apt install", file=sys.stderr)
            elif self.user_can_sudo:
                print(_("Try: %s <deb name>") % "sudo apt install", file=sys.stderr)
            else:
                print(_("Ask your administrator to install one of them."), file=sys.stderr)
            print()
        elif len(packages) > 0 and len(snaps) > 0:
            print()
            print(_("Command '%s' not found, but can be installed as:") % command, file=sys.stderr)
            print()
            for snap in snaps:
                ver=""
                if snap[2]:
                    ver = " (%s) " % snap[2]
                print(" snap %s%s" % (snap[0], ver), file=sys.stderr)
            for package in packages:
                ver=""
                if package[2]:
                    ver = " (%s) " % package[1]
                print(" deb %s%s" % (package[0], ver), file=sys.stderr)
            print()
            print(_("See 'snap info %s' for additional versions.") % command)
            print()
                
        return len(packages) > 0 or len(snaps) > 0
